import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

DEFAULT_RETRY_STRATEGY = Retry(
    total=4,
    backoff_factor=1.5,
    status_forcelist=[429, 500, 502, 503, 504],
)

DEFAULT_TIMEOUT = (10.0, 100.0)


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(
        self, timeout=DEFAULT_TIMEOUT, max_retries=DEFAULT_RETRY_STRATEGY, **kwargs
    ):
        self.timeout = timeout
        super().__init__(max_retries=max_retries, **kwargs)

    def send(self, request, **kwargs):
        if kwargs.get("timeout") is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


HTTP_SESSION_WITH_TIMEOUT_AND_RETRY = requests.Session()
_adapter = TimeoutHTTPAdapter()
HTTP_SESSION_WITH_TIMEOUT_AND_RETRY.mount("http://", _adapter)
HTTP_SESSION_WITH_TIMEOUT_AND_RETRY.mount("https://", _adapter)



import logging

logger = logging.getLogger("bluequbit-python-sdk")


class BQBaseException(Exception):
    def __init__(
        self, status_code="Unknown Status Code", message="Unknown Error Message"
    ):
        self.status_code = status_code
        self.message = message
        super().__init__(status_code, message)


class BQUnauthorizedException(BQBaseException):
    def __init__(self):
        super().__init__(
            401,
            "Wrong API key. You can find your API key once you log in to https://dev.app.bluequbit.io",
        )


class BQStatevectorTooLarge(BQBaseException):
    def __init__(self, num_qubits):
        super().__init__(
            0,
            f"Statevector is too large for {num_qubits} qubits. Use .get_counts() instead.",
        )



# Mypy; for the `|` operator purpose
# Remove this __future__ import once the oldest supported Python is 3.10
import logging

from . import exceptions
from .bluequbit_client import BQClient
from .estimate_result import EstimateResult
from .job_result import JobResult
from .version import __version__

formatter = logging.Formatter(fmt="BQ-PYTHON-SDK - %(levelname)s - %(message)s")
# formatter = logging.Formatter(fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s')

handler = logging.StreamHandler()
handler.setFormatter(formatter)

logger = logging.getLogger("bluequbit-python-sdk")
logger.setLevel(logging.INFO)
logger.addHandler(handler)



import datetime
from io import BytesIO

import numpy as np
from dateutil.parser import parse

from .exceptions import BQStatevectorTooLarge
from .http_adapter import HTTP_SESSION_WITH_TIMEOUT_AND_RETRY


class JobResult:
    """This class contains information from a job run on quantum hardware or
    quantum simulators. Mainly it contains the resulting statevector from the
    run. It might contain only partial information, such as job ID,
    when :func:`BQClient.run` is called with ``asynchronous=True``."""

    def __init__(self, data):
        #: str: job id
        self.job_id = data.get("job_id")

        #: str: job run status, can be one of ``"FAILED_VALIDATION"`` | ``"PENDING"`` | ``"QUEUED"`` | ``"RUNNING"`` | ``"TERMINATED"`` | ``"CANCELED"`` | ``"NOT_ENOUGH_FUNDS"`` | ``"COMPLETED"``
        self.run_status = data.get("run_status")

        #: str: job name
        self.job_name = data.get("job_name")

        #: int: estimated runtime in milliseconds
        self.estimate_ms = data.get("estimated_runtime_ms")

        #: float: estimated cost in US dollars
        self.estimate_cost = data.get("estimated_cost")
        if self.estimate_cost is not None:
            self.estimate_cost /= 100.0

        #: datetime: job creation date in UTC timezone
        self.created_on = data.get("created_on")
        if self.created_on is not None:
            self.created_on = parse(self.created_on)

        self._results_path = data.get("results_path")
        self.top_100_results = data.get("top_100_results")

        #: int: number of qubits
        self.num_qubits = data.get("num_qubits")
        self.tags = data.get("tags")

        #: datetime: job enqueue time in UTC timezone
        self.queue_start = data.get("queue_start")
        if self.queue_start is not None:
            self.queue_start = parse(self.queue_start)

        #: datetime: job run start time in UTC timezone
        self.run_start = data.get("run_start")
        if self.run_start is not None:
            self.run_start = parse(self.run_start)

        #: datetime: job run end time in UTC timezone
        self.run_end = data.get("run_end")
        if self.run_end is not None:
            self.run_end = parse(self.run_end)

        #: int: time spent in queue in milliseconds
        self.queue_time_ms = None
        if self.queue_start is not None and self.run_start is not None:
            self.queue_time_ms = int(
                (self.run_start - self.queue_start) / datetime.timedelta(milliseconds=1)
            )

        #: int: job runtime in milliseconds
        self.run_time_ms = None
        if self.run_start is not None and self.run_end is not None:
            self.run_time_ms = round(
                (self.run_end - self.run_start) / datetime.timedelta(milliseconds=1)
            )

        #: str: error message if failed
        self.error_message = data.get("error_message")

        #: float: job cost in US dollars
        self.cost = data.get("cost")
        if self.cost is not None:
            self.cost /= 100.0
        self._metadata = None
        self._statevector = None

    def get_statevector(self) -> np.array:
        """Return statevector of the job. If the statevector is too large then throws exception.

        :rtype: NumPy array
        """

        if self._statevector is None:
            response = HTTP_SESSION_WITH_TIMEOUT_AND_RETRY.get(
                self._results_path + "statevector.txt"
            )
            if response.ok:
                data = response.content
                self._statevector = np.loadtxt(BytesIO(data), dtype=np.complex_)
        if not self._statevector is None:
            return self._statevector
        else:
            raise BQStatevectorTooLarge(self.num_qubits)

    def get_counts(self) -> dict:
        if self._metadata is None:
            response = HTTP_SESSION_WITH_TIMEOUT_AND_RETRY.get(
                self._results_path + "metadata.json"
            )
            self._metadata = response.json()
        if "counts" not in self._metadata:
            raise RuntimeError("The job result metadata doesn't contain counts.")
        return self._metadata["counts"]

    @property
    def ok(self) -> bool:
        """``True``, if job's current run status is ``"COMPLETED"``,
        else ``False``."""

        return self.run_status == "COMPLETED"

    def __str__(self):
        repr = f"Job ID: {self.job_id}"
        if self.job_name is not None:
            repr += f", name: {self.job_name}"
        repr += f", run status: {self.run_status}, created on: {self.created_on}"
        if (
            self.run_status in ["PENDING", "QUEUED", "RUNNING"]
            and self.estimate_ms is not None
        ):
            repr += f", estimated runtime: {self.estimate_ms} ms"
            repr += f", estimated cost: ${self.estimate_cost:.2f}"
        if self.queue_time_ms is not None:
            repr += f", queue time: {self.queue_time_ms} ms"
        if self.run_time_ms is not None:
            repr += f", run time: {self.run_time_ms} ms"
        if self.cost is not None:
            repr += f", cost: ${self.cost:.2f}"
        if self.num_qubits is not None:
            repr += f", num qubits: {self.num_qubits}"
        if self.error_message is not None:
            repr += f", error_message: {self.error_message}"
        return repr



# Mypy; for the `|` operator purpose
# Remove this __future__ import once the oldest supported Python is 3.10
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

import packaging.version

from .api import jobs
from .backend_connection import BackendConnection
from .estimate_result import EstimateResult
from .http_adapter import HTTP_SESSION_WITH_TIMEOUT_AND_RETRY
from .job_result import JobResult
from .version import __version__


def _check_version():
    local_version = packaging.version.parse(__version__)
    if local_version.is_prerelease:
        logger.warning(
            "Development version %s of BlueQubit Python SDK is being used", __version__
        )
    req = HTTP_SESSION_WITH_TIMEOUT_AND_RETRY.get(
        "https://pypi.python.org/pypi/bluequbit/json", timeout=2.0
    )
    if not req.ok:
        return

    # find max version on PyPI
    releases = req.json().get("releases", [])
    pip_version = packaging.version.parse("0")
    for release in releases:
        ver = packaging.version.parse(release)
        if not ver.is_prerelease or local_version.is_prerelease:
            pip_version = max(pip_version, ver)

    if pip_version.major > local_version.major:
        logger.warning(
            "There is a major upgrade of BlueQubit Python SDK available on PyPI. We recommend upgrading. Run 'pip install --upgrade bluequbit' to upgrade from your version %s to %s.",
            local_version,
            pip_version,
        )
    elif pip_version > local_version:
        logger.info(
            "There is a newer version of BlueQubit Python SDK available on PyPI. Run 'pip install --upgrade bluequbit' to upgrade from your version %s to %s.",
            local_version,
            pip_version,
        )


if TYPE_CHECKING:
    import datetime
    from pathlib import Path

# TODO this requires imports of actual quantum libraries for proper type
# checking.
CircuitT = Any

logger = logging.getLogger("bluequbit-python-sdk")


class BQClient:
    """Client for managing jobs on BlueQubit platform.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in default configuration file ``$HOME/.config/bluequbit/config.json``.
                      If not ``None``, the token will also be saved in the same
                      default configuration file.
    """

    def __init__(self, api_token: str | None = None):
        if os.environ.get("BLUEQUBIT_TESTING") is None:
            try:
                _check_version()
            except:
                pass
        self._backend_connection = BackendConnection(api_token)

    def estimate(
        self,
        circuit: CircuitT,
    ) -> EstimateResult:
        """Estimate job runtime

        :param circuit: quantum circuit
        :type circuit: Cirq, Qiskit, circuit
        :return: estimate result metadata
        """
        return EstimateResult(
            jobs.estimate_job_runtime(self._backend_connection, circuit)
        )

    def run(
        self,
        circuit: CircuitT,
        asynchronous: bool = False,
        job_name: str | None = None,
    ) -> JobResult:
        """Submit a job to run on BlueQubit platform

        :param circuit: quantum circuit
        :type circuit: Cirq, Qiskit, circuit
        :param asynchronous: if set to ``False``, wait for job completion before
                             returning. If set to ``True``, return immediately
        :param job_name: customizable job name
        :return: job metadata
        """
        submitted_job = JobResult(
            jobs.submit_job(self._backend_connection, circuit, job_name)
        )
        if submitted_job.run_status == "FAILED_VALIDATION":
            logger.error(submitted_job)
        else:
            logger.info(f"Submitted: {submitted_job}")
            if not asynchronous:
                return self.wait(submitted_job.job_id)
        return submitted_job

    def wait(self, job_id: str) -> JobResult:
        """Wait for job completion

        :param job_id: job ID that can be found as property of :class:`JobResult` metadata
                       of :func:`~run` method
        :return: job metadata
        """
        return JobResult(jobs.wait_for_job(self._backend_connection, job_id))

    def get(self, job_id: str) -> JobResult:
        """Get current metadata of job

        :param job_id: job ID that can be found as property of :class:`JobResult` metadata
                       of :func:`~run` method
        :return: job metadata
        """
        job_result = jobs.get(self._backend_connection, job_id)
        jr = JobResult(job_result)
        # logger.info(jr)
        return jr

    def cancel(
        self,
        job_id: str,
        asynchronous: bool = False,
    ) -> JobResult:
        """Submit job cancel request

        :param job_id: job ID that can be found as property of :class:`JobResult` metadata
                       of :func:`run` method
        :param asynchronous: if set to ``False``, wait for job cancellation before
                             returning. If set to ``True``, return immediately
        :return: job metadata
        """
        jobs.cancel_job(self._backend_connection, job_id)
        if not asynchronous:
            return self.wait(job_id)
        else:
            return JobResult({"job_id": job_id})

    def search(
        self,
        run_status: str | None = None,
        created_later_than: str | datetime.datetime | None = None,
    ) -> list[JobResult]:
        """Search jobs

        :param run_status: if not ``None``, run status of jobs to filter.
                           Can be one of ``"FAILED_VALIDATION"`` | ``"PENDING"`` |
                           ``"QUEUED"`` | ``"RUNNING"`` | ``"TERMINATED"`` | ``"CANCELED"`` |
                           ``"NOT_ENOUGH_FUNDS"`` | ``"COMPLETED"``

        :param created_later_than: if not ``None``, filter by latest job creation datetime.
                                   Please add timezone for clarity, otherwise UTC
                                   will be assumed

        :return: metadata of jobs
        """
        job_results = jobs.search_jobs(
            self._backend_connection, run_status, created_later_than
        )
        return [JobResult(r) for r in job_results["data"]]



import json
import logging
import os
from pathlib import Path

import requests
import requests_toolbelt
import urllib3

from .exceptions import BQBaseException, BQUnauthorizedException
from .http_adapter import TimeoutHTTPAdapter
from .version import __version__

logger = logging.getLogger("bluequbit-python-sdk")


class BackendConnection:
    def __init__(self, api_token=None):
        config_dir = Path.home() / ".config" / "bluequbit"
        config_location = config_dir / "config.json"

        if api_token is None:
            if not config_location.is_file():
                raise BQBaseException(
                    0,
                    "Please specify an API key to BQClient(). You can find your API key once you log in to https://dev.app.bluequbit.io",
                )
            else:
                with open(config_location, encoding="utf-8") as f:
                    self._api_config = json.load(f)
        else:
            api_config = {
                "token": api_token,
                "main_endpoint": "https://dev.app.bluequbit.io/api/v1",
                "ssl_verify": True,
            }
            self._api_config = api_config
            config_dir.mkdir(exist_ok=True, parents=True)
            self.update_config_file(api_config, config_location)

        try:
            self._token = self._api_config["token"]
        except KeyError:
            raise BQBaseException(
                0,
                "Incorrect config file: 'token' key is not present in the config file "
                + str(config_location),
            )

        try:
            self._default_headers = {"Authorization": f"SDK {self._token}"}
            self._default_headers["User-Agent"] = requests_toolbelt.user_agent(
                "bluequbit", __version__
            )

            self._main_endpoint = "https://dev.app.bluequbit.io/api/v1"
            if "main_endpoint" in self._api_config:
                self._main_endpoint = self._api_config["main_endpoint"]
            self._verify = True
            if "ssl_verify" in self._api_config:
                self._verify = self._api_config["ssl_verify"]
            self._session = None
            self._authenticated = True
            response = self.send_request(
                req_type="GET",
                path="/jobs",
                params={"limit": 1},
            )
            if not self._verify:
                urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

            if not response.ok:
                self._authenticated = False
                self._session = None
                if response.status_code == 401:
                    raise BQUnauthorizedException()
                raise BQBaseException(0, "Couldn't reach BlueQubit " + response.text)
        except:
            self._authenticated = False
            self._session = None
            raise

    @staticmethod
    def update_config_file(content, config_location):
        with open(config_location, "w", encoding="utf-8") as f:
            json.dump(content, f, indent=4)
        logger.info("Configuration file %s successfully updated.", config_location)

    def send_request(self, req_type, path, params=None, json_req=None):
        if not self._authenticated:
            raise BQBaseException(
                0,
                "BQClient was not initialized. Please provide correct API token to BQClient(<token>).",
            )
        url = self._main_endpoint + path

        if params is not None:
            for key, value in params.items():
                if isinstance(value, str):
                    params[key] = value.replace("\\", "\\\\")

        req = requests.Request(method=req_type, url=url, json=json_req, params=params)
        if self._session is None:
            self._session = self._create_session()
        prepared = self._session.prepare_request(req)
        resp = self._session.send(request=prepared, verify=self._verify)
        return resp

    def _create_session(self):
        session = requests.Session()
        adapter = TimeoutHTTPAdapter()
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        session.headers = self._default_headers

        if "BLUEQUBIT_DEBUG" in os.environ:
            session.hooks["response"].append(_log_requests)

        return session


if "BLUEQUBIT_DEBUG" in os.environ:
    from requests_toolbelt.utils import dump

    def _log_requests(response, *args, **kwargs):
        data = dump.dump_all(response)
        _log_requests.response_len += len(data)
        logger.info("HTTP %s %s ", response.request.url, _log_requests.response_len)

    _log_requests.response_len = 0
